"""Top-level package for Microbial Interaction Database Pipeline."""

__author__ = """Dileep Kishore"""
__email__ = "dkishore@bu.edu"
__version__ = "0.7.1"

from .main import Lineage, Network, NetworkGroup, Otu
